/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill;

import java.io.IOException;

/**
 *  Interface <CODE>AntView</CODE> contains the set of methods which are
 *  accessible to ants when visiting nests. These methods enable ants to 
 *  cooperate with nests in a controlled way. For example, ants may
 *  retrieve resources from or add resources to a nest; obtain routing 
 *  information related to other nests; move to other nests, and so on.
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision: 1.16 $
 */
public interface AntView
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for requesting general information
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns the identifier of the current nest hosting the
   *  ant using this <CODE>AntView</CODE>.
   */
  public NestId getNestId();

  
  /**
   *  Returns the identifier of the last nest visisted by this
   *  ant.
   */
  public NestId getLastVisited();
  
  /**
   */
  public int getTTL();

  /**
   *  Returns the ant which is associated to this <CODE>AntView</CODE>. Normally not
   *  used by ants, as they may access the same information using
   *  <CODE>this</CODE>.
   */
  public Ant getAnt();

  
  /**
   *  Creates a new ant identifier.
   */
  public AntId createAntId();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for ant movement and creation
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Add a new ant to this nest. For example, this method
   *  could be used to create a backward ant for returning
   *  information to the requestor nest.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *  
   *  @param ant the ant to be added to the current nest
   */
  public void addAnt(Ant ant);

  /**
   *  Method invoked by an ant in order to be sent to another 
   *  <CODE>Nest</CODE>. It is possible that particular 
   *  <CODE>AntView</CODE> implementation limit the operations 
   *  which can be performed after a move operation. For example, 
   *  movement to more than one nest (i.e., duplication) may not
   *  be enabled.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *
   *  @param id the identifier of the destination nest
   *  @throws IOException raised if the move operation fails,
   *    for example due to the unreachability of the nest
   */
  public void move(NestId id)
    throws IOException;
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for neighbour management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Ants invoke this method to add the identifier of a new neighbour
   *  nest to the set of neighbours known to the current nest. Depending
   *  on the implementation, this operation may require the establishment
   *  of a connection to the new nest. In this case, an <CODE>IOException</CODE> 
   *  may be raised in the case a connection cannot be established.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   *  
   *  @throws IOException if a connection with the nest cannot be established.
   */
  public void addNeighbour(NestId id)
    throws IOException;

  /**
   *  Ants invoke this method to remove a neighbor nest from the list of
   *  neighbors known to this nest. Depending on the implementation, this 
   *  operation may require the closure of a previously created nest connection.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   */
  public void removeNeighbour(NestId id);
  
  /**
   *  Ants invoke this method to obtain the list of nest neighbors
   *  known to the current nest.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  current nest hosting the ant. Please refer to the documentation of
   *  the nest implementation for more details.
   */
  public Neighbour[] getNeighbours();

    
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for resource management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a reference to the resource manager identified by the given name.
   *  Returns null if no resource manager is associated with the given name
   *  to this ant.
   */
  public ResourceManager getResourceManager(String name);
  
  /**
   *  Returns a reference to the resource manager identified by the given name,
   *  if present. Otherwise, creates a new resource manager, using <CODE>name</CODE> as 
   *  lookup identifier in the factory object to obtain the current configuration.
   *  
   *  @throws ResourceManager this exception is thrown if, for some reason, 
   *    it is impossible to create a resource manager
   */
  public ResourceManager createResourceManager(String name)
    throws ResourceManagerException;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods for resource management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  This method is invoked by ants to notify the nest that
   *  the ant associated to this <CODE>AntView</CODE>, created to satisfy
   *  a request, has found the specified set of resources.
   *
   *  @param resources  the resources satisfying the request
   */
  public boolean returnResources(RequestId rid, Resource[] resources);
  

} // END AntView
