/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.factory;

import java.io.Serializable;
import java.lang.reflect.*;
import java.util.*;

/**
 *  A generic factory for creating object instances of a class
 *  matching a particular interface and possibly taking arguments.
 *  The mapping between an implementing class and an interface,
 *  and also the arguments to be used to instantiate the class
 *  are all read from a configuration file.
 *
 *  @author Hein Meling
 *  @version $Revision: 1.14 $
 */
public class Factory implements Serializable
{
  /////////////////////////////////////////////////////////////////////////////
  // Constants - Error message
  /////////////////////////////////////////////////////////////////////////////

  private static final String errorMsg = "Error in configuration file: ";

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /*  Map of interfaces mapping to class, args pairs. */
  private HashMap componentMap = new HashMap();

  /*  Map of configuration names mapping to configuration values */
  private Map configMap = new TreeMap();

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Create a parser object and pass this factory object allowing the
   *  parser to place the parsed elements directly into the HashMap
   *  using the factory API method <CODE>put</CODE>.  The parser will
   *  use the provided URLs for the configuration file location.
   *
   *  @param configURLs 
   *    An array of URLs pointing to the configuration files to parse.
   *  @throws ConfigFormatException 
   *    Thrown when there is a problem parsing or validating the content 
   *    of the system configuration file.
   *  @throws IOException 
   *    Thrown when one of the urls cannot be accessed due to an I/O
   *    problem.
   */
  public Factory(String[] configURLs)
	throws ConfigFormatException, java.io.IOException
  {
	ConfigParser cp = new ConfigParser(this);
	for (int i=0; i < configURLs.length; i++) {
	  cp.read(configURLs[i]);
	}
  }

  /**
   *  Create a parser object and pass this factory object allowing the
   *  parser to place the parsed elements directly into the HashMap
   *  using the factory API method <CODE>put</CODE>.  The parser will
   *  use the default configuration file location.
   *
   *  @throws ConfigFormatException 
   *    Thrown when there is a problem parsing or validating the content 
   *    of the system configuration file.
   *  @throws IOException 
   *    Thrown when one of the urls cannot be accessed due to an I/O
   *    problem.
   */
  public Factory()
	throws ConfigFormatException, java.io.IOException
  {
	ConfigParser cp = new ConfigParser(this);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Factory interface methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  The <CODE>get</CODE> method is used to obtain an instantiation of         
   *  a class implementing the requested interface.
   *
   *  @param cl The interface used as mapping key.
   *  @return   An object instance implementing the specified interface.
   *
   *  @throws ConfigFormatException  Thrown if specified classes and interfaces
   *     cannot be found.
   *  @throws ConstructorException Thrown if the constructor reflectively invoked
   *    by get raises an exception
   */
  public Object get(Class cl)
	throws ConfigFormatException, ConstructorException
  {
	return get(cl.getName());
  }

  /**
   *  The <CODE>get</CODE> method is used to obtain an instantiation of
   *  a class implementing the requested component.
   *
   *  @param name  The name of the component used as a mapping key.
   *  @return       An object instance implementing the specified component.
   *
   *  @throws ConfigFormatException  Thrown if specified classes cannot be found.
   *  @throws ConstructorException Thrown if the constructor reflectively invoked
   *    raises an exception
   */
  public Object get(String name)
	throws ConfigFormatException, ConstructorException
  {
	// System.out.println("Factory: " + name);
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();

	// Obtain the class name associated to the specified component name.
	String classname = (String) componentMap.get(name);
	if (classname == null) {
	  throw new ConfigFormatException(errorMsg + "component not defined: " + name);
	}

	try {
	  /* Get a Class object representing the specified class. */
	  Class clazz = Class.forName(classname);
	  
	  // Obtain a constructor with a single argument of type Factory
	  Class[] types = new Class[] { Factory.class, String.class };
	  Constructor c = clazz.getConstructor(types);
	  
	  // Invoke constructor with the specified factory
	  Object[] args = new Object[] { this, name };
	  return (Object) c.newInstance(args);
	} catch (ClassNotFoundException exception) {
	  throw new ConfigFormatException(errorMsg
		+ "No matching class " + classname + " was found.");
	} catch (NoSuchMethodException exception) {
	  throw new ConfigFormatException(errorMsg
		+ "Found no matching constructor for class " + classname);
	} catch (InstantiationException exception) {
	  throw new ConfigFormatException(errorMsg
		+ "The specified class " + classname
		+ " is an interface or abstract class.");
	} catch (IllegalAccessException exception) {
	  throw new ConfigFormatException(errorMsg
		+ "The specified class " + classname + " is not accessible.");
	} catch (InvocationTargetException exception) {
	  Throwable e = exception.getTargetException();
	  while (e instanceof ConstructorException) {
		e = ((ConstructorException) e).getTargetException();
	  }
	  throw new ConstructorException(e);
	}
  }


  /**
   *  Method used to update the mapping between a component name and the
   *  class implementing it.
   *
   *  @param name The interface name used as mapping key.
   *  @param clazz  The class name implementing the above interface.
   *
   */
  public void put(String name, String clazz) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	componentMap.put(name, clazz);
  }

  /**
   *  Method used to insert or update a configuration value
   *  associated to a configuration name.
   *
   *  @param name The name associated to the value.
   *  @param value The value to be inserted.
   */
  public void putValue(String name, Object value) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo info = (ArgInfo) configMap.get(name);
	if (info != null) {
	  info.value = value;
	} else {
	  info = new ArgInfo(value);
	  configMap.put(name, info);
	}
  }

  /**
   *  Method used to insert or update an integer configuration value
   *  associated to a configuration name.
   *
   *  @param name The name associated to the value.
   *  @param value The value to be inserted.
   */
  public void putValue(String name, int value) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  pair.value = new Integer(value);
	} else {
	  pair = new ArgInfo(new Integer(value));
	  configMap.put(name, pair);
	}
  }

  /**
   *  Method used to insert or update a double configuration value
   *  associated to a configuration name.
   *
   *  @param name The name associated to the value.
   *  @param value The value to be inserted.
   */
  public void putValue(String name, double value) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  pair.value = new Double(value);
	} else {
	  pair = new ArgInfo(new Double(value));
	  configMap.put(name, pair);
	}
  }

  /**
   *  Method used to insert or update a double configuration value
   *  associated to a configuration name.
   *
   *  @param name The name associated to the value.
   *  @param value The value to be inserted.
   */
  public void putValue(String name, float value) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  pair.value = new Float(value);
	} else {
	  pair = new ArgInfo(new Float(value));
	  configMap.put(name, pair);
	}
  }

  /**
   *  Method used to obtain the value associated to a 
   *  particular configuration parameter
   *
   *  @param name The name identifying the configuration parameter
   */
  public Object getValue(String name) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  return pair.value;
	} else {
	  throw new IllegalArgumentException("No value identified by " + name);
	}
  }

  /**
   *  Method used to obtain the String value associated to a 
   *  particular configuration parameter
   *
   *  @param name The name identifying the configuration parameter
   */
  public String stringValue(String name) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  return (String) pair.value;
	} else {
	  throw new IllegalArgumentException("No value identified by " + name);
	}
  }

  /**
   *  Method used to obtain the boolean value associated to a 
   *  particular configuration parameter
   *
   *  @param name The name identifying the configuration parameter
   */
  public boolean booleanValue(String name) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  return ((Boolean) pair.value).booleanValue();
	} else {
	  throw new IllegalArgumentException("No value identified by " + name);
	}
  }
  /**
   *  Method used to obtain the integer value associated to a 
   *  particular configuration parameter
   *
   *  @param name The name identifying the configuration parameter
   */
  public int intValue(String name) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  return ((Number) pair.value).intValue();
	} else {
	  throw new IllegalArgumentException("No value identified by " + name);
	}
  }

  /**
   *  Method used to obtain the double value associated to a 
   *  particular configuration parameter
   *
   *  @param name The name identifying the configuration parameter
   */
  public double doubleValue(String name) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  return ((Number) pair.value).doubleValue();
	} else {
	  throw new IllegalArgumentException("No value identified by " + name);
	}
  }

  /**
   *  Method used to obtain the double value associated to a 
   *  particular configuration parameter
   *
   *  @param name The name identifying the configuration parameter
   */
  public float floatValue(String name) 
  {
	// Convert the string to lower case, to simplify reading
	name = name.toLowerCase();
	ArgInfo pair = (ArgInfo) configMap.get(name);
	if (pair != null) {
	  return ((Number) pair.value).floatValue();
	} else {
	  throw new IllegalArgumentException("No value identified by " + name);
	}
  }

  public String toString()
  {
	StringBuffer buffer = new StringBuffer();
	buffer.append("Factory: \n");
	Iterator it = componentMap.keySet().iterator();
	while (it.hasNext()) {
	  String name = (String) it.next();
	  String clazz = (String) componentMap.get(name);
	  buffer.append("- ");
	  buffer.append(name);
	  buffer.append(": ");
	  buffer.append(clazz);
	  buffer.append("\n");
	}
	it = configMap.keySet().iterator();
	while (it.hasNext()) {
	  String name = (String) it.next();
	  ArgInfo arg = (ArgInfo) configMap.get(name);
	  buffer.append("- ");
	  print(buffer, null, name, arg.value);
	  buffer.append("\n");
	}
	return buffer.toString();
  }

  private void print(StringBuffer buffer, Class type, String name, Object value)
  {
	if (type != Factory.class) {
	  buffer.append(name);
	  buffer.append(" = ");
	  if (type == String.class)
		buffer.append('"');
	  buffer.append(value);
	  if (type == String.class)
		buffer.append('"');
	} else {
	  buffer.append("Factory");
	}
  }
  
  
} // END Factory
