/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.nest;

import anthill.*;
import anthill.util.Clearable;

/**
 *  <CODE>Nest</CODE> is the interface used by nest users
 *  to communicate with nests. The basic methods contained in this 
 *  interface enable the user to search resources, to insert 
 *  resources in the Anthill network and to add knowledge
 *  about other nests.
 *
 *  @author Alberto Montresor
 *  @version $Revision: 1.10 $
 */
public interface Nest
extends Clearable
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Identifier used in the configuration files */
  public static final String NAME = "Nest";
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Nest management
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the identifier of this nest
   */
  public NestId getNestId();

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Resource management
  ////////////////////////////////////////////////////////////////////////////////////////////
  

  /**
   *  Returns a reference to the resource manager associated to the specified
   *  ant species and identified by the given name. Different ant species may
   *  be associated to different resource managers; parameter <CODE>species</CODE>
   *  is used to discriminate among ant species. If no resource manager is
   *  associated to <CODE>species</CODE> and is identified by <CODE>name</CODE>,
   *  the method returns <CODE>null</CODE>.
   *  
   *  @param name the name of the manager to obtain a reference to
   *  @param species the name of the ant species to which the manager is
   *    associated
   *  @return a reference to the requested resource manager, or <CODE>null</CODE>
   *    if no such resource manager is found.
   */
  public ResourceManager getResourceManager(String name, String species);

  /**
   *  Returns a reference to the resource manager associated to the specified
   *  ant species and identified by the given name. Different ant species may
   *  be associated to different resource managers; parameter <CODE>species</CODE>
   *  is used to discriminate among ant species. If no resource manager is
   *  associated to <CODE>species</CODE> and is identified by <CODE>name</CODE>,
   *  a new resource manager is created, using <CODE>name</CODE> as 
   *  lookup identifier in the factory object to obtain the current configuration.
   *  
   *  @param name the name of the manager to obtain a reference to
   *  @param species the name of the ant species to which the manager is
   *    associated
   *  @return a reference to the requested resource manager, or <CODE>null</CODE>
   *    if no such resource manager is found.
   *  @throws ResourceManager this exception is thrown if, for some reason, 
   *    it is impossible to create a resource manager
   */
  public ResourceManager createResourceManager(String name, String species)
    throws ResourceManagerException;

  /**
   *  Accepts a resource request originated by the user of this nest.
   *  The interpretation of the request object is left to the ant(s)
   *  created to satisfy the response. This method returns immediately.
   *  When the request is completed, one or more <CODE>Resource</CODE> 
   *  object will be returned to the specified resource listener.
   *
   *  @param rid the request identifier, created and managed by the
   *    invoker of this method.
   *  @param request the request to be satisfied
   *  @param listener The listener object which will be notified
   *    when the request is completed.
   */
  public void request(RequestId rid, Object request, ResponseListener listener);
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Neighbour list management
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Adds the identifier of a new neighbour nest to the set of 
   *  neighbours known to this nest, on behalf of the nest user. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  gate object contained in the nest. Please refer to the documentation 
   *  of the resource storage implementation for more details.
   *
   *  @param nid the identifier of the neighbour to be added
   */
  public void addClientNeighbour(NestId nid);

  /**
   *  Adds the identifier of a new neighbour nest to the set of 
   *  neighbours known to this nest, on behalf of an ant visiting
   *  the nest.
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  gate object contained in the nest. Please refer to the documentation 
   *  of the resource storage implementation for more details.
   *
   *  @param nid the identifier of the neighbour to be added
   */
  public void addAntNeighbour(NestId nid);

  /**
   *  Removes the identifier of a neighbour nest from the set of 
   *  neighbours known to this Nest. 
   *  <p>
   *  The exact behavior of this method depends on the implementation of the
   *  gate object contained in the nest. Please refer to the documentation 
   *  of the resource storage implementation for more details.
   *
   *  @param nid the identifier of the neighbour to be removed
   */
  public void removeNeighbour(NestId nid);
  
  /**
   * Returns information about the neighbours known to this nest.
   */
  public Neighbour[] getNeighbours();

  
} // END Nest
