/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.storage.impl;

import java.util.*;

import anthill.Resource;
import anthill.factory.Factory;
import anthill.storage.*;

/**
 *  The <CODE>MemoryStorage</CODE> is a simple implementation of
 *  the generic storage which keeps resources in memory.
 *  <p>
 *  Implementation note: currently, no check on the actual memory
 *  occupation of an object is performed; this implementation
 *  use the size() field of the object to obtain
 *  the length of the object. Future implementations for real
 *  deployment will include this additional check.
 *  
 *  @author Alberto Montresor
 *  @version $Revision: 1.8 $
 */
public class MemoryStorage
implements GenericStorage
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameter identifiers
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** 
   *  The identifier of the configuration parameter 
   */
  public static final String PAR_CAPACITY = "Capacity";
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Maximum capacity of the storage */
  private long capacity;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory used to instantiate objects */
  private Factory factory;
  
  /** Name used to identify this component in the factory */
  private String name;
  
  /** Remaining capacity */
  private long remainingCapacity;
  
  /** Hashmap used to map keys to resources */
  private HashMap map;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructors
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Constructor for a memory capacity object
   *
   *  @param factory the factory to be used to read configuration arguments
   *    and to create other configurable objects
   *  @param name the name that is associated to this instance in the
   *    configuration file
   */
  public MemoryStorage(Factory factory, String name)
  {
    this.factory = factory;
    this.name = name;
    clear();
  }
  
  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    /*
     *  Read the maximum capacity from the factory
     */
    capacity = factory.intValue(name + "." + PAR_CAPACITY);
    remainingCapacity = capacity;
    if (map == null) {
      map = new HashMap();
    } else {
      map.clear();
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns true if the resource associated to <CODE>key</CODE> is contained in
   *  the storage, false otherwise.
   *
   *  @param key  the key to be searched
   *  @return true if the resource associated to <CODE>key</CODE> is contained in
   *    the storage
   */
  public boolean containsResource(Object key)
  {
    return map.containsKey(key);
  }


  /**
   *  Returns the resource identified by the given key. If the key
   *  is not associated with any resource, a null value is returned.
   */
  public Resource getResource(Object key)
  {
    return (Resource) map.get(key);
  }

  /**
   *  Returns the complete set of resources stored in this storage object
   *  using an iterator.
   */
  public Iterator iterator()
  {
    return map.values().iterator();
  }

  /**
   *  Adds a resource to the storage and associates it to the specified
   *  key. If the remaining capacity of the storage is not sufficient to
   *  store the resource, a <CODE>CapacityException</CODE> is thrown
   *  containing information about the remaining capacity. Method
   *  <CODE>size()</CODE> may be invoked on the resource
   *  to determine the amount of space required by the resource. 
   *  In order to be stored on persistent storage, resource must 
   *  implement the <CODE>Serializable</CODE> interface. 
   *
   *  @param key the key identifying the resource
   *  @param resource the resource to be added
   */
  public void addResource(Object key, Resource resource)
    throws CapacityException
  {
    long len = resource.size();
    if (len > remainingCapacity || len > capacity) {
      throw new CapacityException(capacity, remainingCapacity, len);
    } else {
      removeResource(key);
      remainingCapacity -= len;
      //System.out.println("Called this " + remainingCapacity);
      map.put(key, resource);
    }
  }
    
  
  /**
   *  Removes a resource from the storage.
   *
   *  @param key the identifier of the resource to be removed
   */
  public void removeResource(Object key)
  {
    Resource old = (Resource) map.remove(key);
    if (old != null) 
      remainingCapacity += old.size();
  }
  
  
  /**
   *  Returns the initial capacity of this generic storage.
   */
  public long getCapacity()
  {
    return capacity;
  }

  
  /**
   *  Returns the remaining capacity of this generic storage.
   */
  public long getRemainingCapacity()
  {
    return remainingCapacity;
  }
  
} // END MemoryStorage
