/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package anthill.types;

import java.util.Random;

import anthill.factory.Factory;

/**
 *  In this simple implementation of the <CODE>HashedKeyword</CODE> interface,
 *  keywords are hashed using the <CODE>hashCode</CODE> method of Java string.
 *  The main advantage of this implementation with respect to more robust
 *  hashing functions like SHA are its small memory footprint and its performance.
 *
 *  @author Alberto Montresor (montresor@cs.unibo.it)
 *  @version $Revision: 1.4 $
 */
public class SimpleHashedKeyword 
implements HashedKeyword
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  private static Random random = new Random();

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The plaintext keyword */
  private String keyword;
  
  /** The hash value associated to the keyword */
  private int hashValue;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Default constructor for externalization. 
   */
  public SimpleHashedKeyword(Factory factory, String name)
  {
  }
  
  /**
   *  Constructs a new hashed keyword starting from a plaintext
   *  keyword. If the plaintext keyword is null, a random integer
   *  is generated and it is used as a keyword.
   */
  public SimpleHashedKeyword(String keyword)
  {
    init(keyword);
  }
  

  /**
   *  Initializes a new hashed keyword starting from a plaintext
   *  keyword. If the plaintext keyword is null, a random integer
   *  is generated and it is used as a keyword.
   */
  public void init(String keyword)
  {
    if (keyword == null) {
      hashValue = random.nextInt();
    } else {
      this.keyword = keyword;
      this.hashValue = keyword.hashCode();
    }
  }
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from HashedKeyword
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the keyword associated with this hashed keyword.
   */
  public String getKeyword()
  {
    return keyword;
  }

  /**
   * Returns the bytes representing the hashed value of the keyword.
   */
  public byte[] getBytes()
  {
    byte[] ret = new byte[4];
    // FIXME Complete this implementation
    return ret;
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Object, Comparable, Closeness
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a string representation of the object.
   */
  public String toString()
  {
    StringBuffer buf = new StringBuffer();
    buf.append("[SimpleHK: ");
    buf.append(keyword);
    buf.append(" -> ");
    buf.append(hashValue);
    buf.append("]");
    return buf.toString();
  }

  /**
   *  Check whether this hashed keyword is equal to the specified object.
   * 
   */
  public boolean equals(Object obj)
  {
    
    if (obj != null && obj instanceof SimpleHashedKeyword) {
      return ((SimpleHashedKeyword) obj).hashValue == hashValue;
    } else {
      return false;
    }
  }
  
  /**
   *  Returns a hash code value for the object.
   */
  public int hashCode()
  {
    return hashValue;
  }

  /**
   *  Compares this object with the specified object for order. Returns a 
   *  negative integer, zero, or a positive integer as this object is less 
   *  than, equal to, or greater than the specified object. 
   *  
   *  @param obj the object to be compared
   */
  public int compareTo(Object obj)
    throws ClassCastException
  {
    SimpleHashedKeyword shk = (SimpleHashedKeyword) obj;
    if (shk.hashValue < hashValue) 
        return -1;
    else if (shk.hashValue == hashValue) 
      return 0;
    else
      return 1;
  }

  /**
   *  Compares <code>this</code> object against the provided objects for
   *  closeness.  An implementation of this method must return an
   *  <code>int</code> according to the following rules:
   *
   *  <pre>
   *    -1 is returned if the value of x is considered closer to this object than the value of y.
   *    +1 is returned if the value of y is considered closer to this object than the value of x.
   *     0 is returned if the values of x and y are equally close to this object.
   *  </pre>
   *  
   *  @throws ClassCastException if x or y are not <CODE>SimpleHashedKeyword</CODE> objects
   */
  public int isCloser(Object x, Object y)
  {
    int dx = Math.abs(hashValue - ((SimpleHashedKeyword) x).hashValue);
    int dy = Math.abs(hashValue - ((SimpleHashedKeyword) y).hashValue);
    if (dx < dy)
      return -1;
    else if (dx == dy) 
      return 0;
    else 
      return 1;
  }

} // END SimpleHashedKeyword
