/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package ants.freeant;

import java.io.*;
import java.util.*;

import anthill.*;
import anthill.storage.*;
import anthill.types.*;

/**
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision: 1.6 $
 */
public class Gnutant
implements Ant, Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Constant value identifying the status of the ant when is searching 
   */
  private static final int ST_SEARCH = 0;
  
  /** 
   *  Constant value identifying the status of the ant when is returning to 
   *  the requestor nest 
   */
  private static final int ST_RETURN = 1;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Probability to follow a random path instead of following the pheromone
   *  trail.
   */
  private float explorationProbability;
  
  /**
   *  Number of attempts to move to another nest, selected randomly among the
   *  nests included in this pheromone, using a normal distribution over the
   *  nests addressing the closest keys.
   */
  private int trailAttempts;
  
  /**
   *  
   */
  private double trailDev;
  
  
  /**
   *  Number of attempts to move to another  nest, selected randomly among the
   *  nests included in this pheromone, using a uniform distribution.
   */
  private int explorationAttempts;
  
  
  /**
   *  
   */
  private int maxResources;
  
  /**
   *  
   */
  private int maxProviders;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Initialization fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The identifier of the request */
  private RequestId rid;

  /** The query string associated with this ant */
  private String query;
  
  /** The hash value of the keyword searched by this ant */
  private HashedKeyword key;
  
  /** Originating nest of this ant */
  private NestId requestor;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // State fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Current ant status */
  private int status;

  /** 
   * Utility object used to obtain information about the return path to 
   * the requestor nest.
   */
  ReturnPath returnPath;
  
  /** Resource to be returned */
  HashSet resources;

  /** First nests to provide new info FIXME currently just one */
  ArrayList providers;
  
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  This parameter represents the probability for the requestor identifier
   *  to be substituted with the current nest, in order to provide anonymity.
   */ 
  private int fakeProb;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate a new Gnutant by setting the query to be satisfied and the
   *  hash value to be searched.
   */
  public Gnutant(RequestId rid, String query, HashedKeyword key, NestId requestor, 
    float explorationProbability, int trailAttempts, double trailDev, 
    int explorationAttempts, int maxResources, int maxProviders)
  {
    
    /* Store initialization parameters */
    this.rid = rid;
    this.query = query;
    this.key = key;
    this.requestor = requestor;
    this.returnPath = new ArrayReturnPath();
    
    /* Initialize parameters */
    this.explorationProbability = explorationProbability;
    this.trailAttempts = trailAttempts;
    this.trailDev = trailDev;
    this.explorationAttempts = explorationAttempts;
    this.maxProviders = maxProviders;
    this.maxResources = maxResources;

    /* Set state variables */
    status = ST_SEARCH;
    providers = new ArrayList(maxProviders);
    resources = new HashSet();
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  // Documentation inherited from interface
  public String getSpecies()
  {
    return "Gnutant";
  }

  
  /**
   *  FIXME: add this method to AntView; add the invocation of this
   *  method in the local nest. Note that this method should be used only for
   *  ant configuration, so we may want to disable move and addAnt
   *  in view.
   */
  public void init(AntView view)
  {
    /* */
    returnPath.init(view);
  }
  
  /**
   *  This method encapsulate the computation performed by the
   *  ant. 
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    //Log.println(500, rid + " at " + view.getNestId()); 
    
    
    /* Decides the behavior to follow, based on current ant status */
    switch (status) {
    case ST_SEARCH: 
      doSearch(view);
      break;
    case ST_RETURN:
      doReturn(view);
      break;
    default:
      System.err.println("Unrecognized status");
    }
  }
  
  /**
   *  
   */
  public void doSearch(AntView view)
  {
    /* Obtain the pheromone storage; creates it if no storage is present */
    GnutantPheromone pheromone = 
      (GnutantPheromone) view.getResourceManager(GnutantPheromone.NAME);
    if (pheromone == null) {
      pheromone = (GnutantPheromone) view.createResourceManager(GnutantPheromone.NAME);
      pheromone.init(view, explorationProbability, trailAttempts, trailDev, explorationAttempts); // FIXME Put arguments
    }


    /* Initialize boolean variable needed to understand whether this
     * ant has to start go back or not. The reason for going back are:
     * (i) we have exhausting the space reserved for the path (half the TTL)
     * (ii) the ant has collected enough responses / nests with responses
     * (iii) the ant cannot move to other nests, because of failures.
     * In this first assignment, only the first condition is checked.
     * The other will be checked later.
     */
    boolean mustGoBack = ( view.getTTL() == returnPath.capacity()+1 );
    
    /* Leave back trail info to return to the requestor node 
     * Discover if this node has been already visited. If not,
     * inspects the node.
     */ 
    boolean firsttime = !returnPath.addLast(view, view.getLastVisited());
    if (firsttime) {
      // System.out.print("+");
      /* Search documents in the storages */
      mustGoBack |= checkStorage(view);
    } else {
      // System.out.print("*");
    }
    
    /* If this ant is not forced to going back, we move to the next
     * node using the methods of pheromone. 
     */
    boolean moved = false;
    if (!mustGoBack) {
      if (firsttime) {
        moved = pheromone.moveToNextTrailNest(view, key);
      } else {
        moved = pheromone.moveToNextRandomNest(view);
      }
    }
    
    if (!moved) {
      status = ST_RETURN;
      //System.out.print("+");
      doReturn(view);
    }
  }
  
  /**
   *  
   */
  private boolean checkStorage(AntView view)
  {
    /* Obtain references to local storages */
    QueryStorage docStorage = (QueryStorage) view.createResourceManager(Storage.DOC_STORAGE);
    QueryStorage urlStorage = (QueryStorage) view.createResourceManager(Storage.URL_STORAGE);

    /* Create an empty array list for the new resources */
    HashSet newResources = new HashSet();
    
    /* Add resources found in the doc storage */
    if (docStorage != null) {
      Resource[] docs = docStorage.getResources(query);
      if (docs != null) {
        for (int i=0; i < docs.length; i++) {
          if (resources.add(docs[i]))
            newResources.add(docs[i]);
        }
      }
    }
    
    /* Add resources found in the url storage */
    if (urlStorage != null) {
      Resource[] urls = urlStorage.getResources(query);
      if (urls != null) {
        for (int i=0; i < urls.length; i++) {
          if (resources.add(urls[i]))
            newResources.add(urls[i]);
        }
      }
    }

    /*  
     * If there are new resources, send a ResponseGnutant and (possibly) set 
     * the originating nest
     */
    if (newResources.size() > 0) {

      /* Send response ant */
      Resource[] news = new Resource[newResources.size()];
      newResources.toArray(news);
      ResponseGnutant rgant = new ResponseGnutant(rid, news, requestor);
      view.addAnt(rgant);
      
      /* Add resources to the list of found resources */
      providers.add(view.getNestId());
      for (int i=0; i < news.length && resources.size() < maxResources; i++) 
        resources.add(news[i]);
    }
    return resources.size() >= maxResources || providers.size() >= maxProviders;
  }

  
  /** 
   *   
   */
  private void doReturn(AntView view)
  {
    /* Add found resources to urlStorage */
    UrlStorage urlStorage = (UrlStorage) view.createResourceManager(Storage.URL_STORAGE);
    Iterator iterator = resources.iterator();
    while (iterator.hasNext()) {
      URLResource resource = (URLResource) iterator.next();
      try {
        urlStorage.addResource(resource.getKey(), resource);
      } catch (CapacityException e) { }
    }
    
    /* Updates the routing information for the keyhash it is searching */
    GnutantPheromone pher = 
      (GnutantPheromone) view.createResourceManager(GnutantPheromone.NAME);
    for (int i=0; i < providers.size(); i++) {
      NestId provider = (NestId) providers.get(i);
      if (!provider.equals(view.getNestId())) {
        /* Add the keyhash and the provder nest to the local pheromone table. */
        pher.addKeyhash(key, provider);
      }
    }
    
    
    
    /*  The ant tries to move to the previous nest of the return path.
     * If the value returned by path.removeLast(view) is null, the
     * ant is lost and cannot return to the original nest.
     */
    NestId previous;
    while ( (previous = returnPath.removeLast(view)) != null) {
      /*  Check whether if the ant is trying to move to the same nest, in
       * which case the nest is skipped and we move to the previous nest.
       */
      if (previous != view.getNestId()) {
        try {
          /*  We move to the previous nest in the path; if successful,
           * we may terminate the execution of this algorithm.
           */
          view.move(previous);
          return;
        } catch (Exception e) {
          /*  Do nothing; the previous nest in the list is unreachable; 
           * the ant tries to select another previous nest, if possible
           */
        }
      }
    }
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void writeExternal(ObjectOutput out)
    throws IOException
  {
    // FIXME To be implemented
  }
  
  public void readExternal(ObjectInput in)
  {
    // FIXME To be implemented
  }
  
} // END Gnutant


