/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package ants.freeant;

import java.util.StringTokenizer;

import anthill.*;
import anthill.factory.Factory;
import anthill.types.HashedKeyword;
import anthill.util.Log;

/**
 *
 *  @author Alberto Montresor
 *  @author Hein Meling
 *  @version $Revision: 1.5 $
 */
public class GnutantFactory
implements AntFactory
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The tokens used to split a keyword string. */
  private static final String TOKENS = " _-.";

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameter names
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Parameter name used to read explorationProbability */
  public static final String PAR_EXPLORATIONPROB = "ExplorationProbability";
  
  /** Parameter name used to read trailAttempts */
  public static final String PAR_TRAILATTEMPTS = "TrailAttempts";
  
  /** Parameter name used to read trailDev */
  public static final String PAR_TRAILDEV = "TrailDev";

  /** Parameter name used to read explorationAttempts */
  public static final String PAR_EXPLORATIONATTEMPTS = "ExplorationAttempts";
  
  /** Parameter name used to read trailDev */
  public static final String PAR_MAXRESOURCES = "MaxResources";

  /** Parameter name used to read trailDev */
  public static final String PAR_MAXPROVIDERS = "MaxProviders";
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Probability to follow a random path instead of following the pheromone
   *  trail.
   */
  private float explorationProbability;
  
  /**
   *  Number of attempts to move to another nest, selected randomly among the
   *  nests included in this pheromone, using a normal distribution over the
   *  nests addressing the closest keys.
   */
  private int trailAttempts;
  
  /**
   *  
   */
  private double trailDev;
  
  
  /**
   *  Number of attempts to move to another  nest, selected randomly among the
   *  nests included in this pheromone, using a uniform distribution.
   */
  private int explorationAttempts;
  
  /**
   *  
   */
  private int maxResources;
  
  /**
   *  
   */
  private int maxProviders;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory reference */
  private Factory factory;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Factory for creating <i>Gnutant ant species</i>.
   *  
   *  @param factory the factory to be used to read configuration arguments
   *    and to create other configurable objects
   *  @param name the name that is associated to this instance in the
   *    configuration file
   */
  public GnutantFactory(Factory factory, String name)
  {
    this.factory = factory;
    clear();
  }

  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    explorationProbability = 
      factory.floatValue(NAME + "." + PAR_EXPLORATIONPROB);
    trailAttempts = 
      factory.intValue(NAME + "." + PAR_TRAILATTEMPTS);
    trailDev = 
      factory.doubleValue(NAME + "." + PAR_TRAILDEV);
    explorationAttempts = 
      factory.intValue(NAME + "." + PAR_EXPLORATIONATTEMPTS);
    maxResources = 
      factory.intValue(NAME + "." + PAR_MAXRESOURCES);
    maxProviders = 
      factory.intValue(NAME + "." + PAR_MAXPROVIDERS);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the AntFactory interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *   Returns a set of ants for a query. Each ant will take care of
   *   the research of a distinct token of the query string. 
   */
  public Ant[] getAnts(RequestId rid, Object request, NestId requestor)
  {
    if (request instanceof SearchRequest) {
      return getSearchAnts(rid, (SearchRequest) request, requestor);
    } else if (request instanceof InsertRequest) {
      return getInsertAnts(rid, (InsertRequest) request, requestor);
    } else {
      Log.println(600, "RequestObject: " + request);
      throw new IllegalArgumentException("Request object not recognized.");
    }
  }
    
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Returns a set of search ants, one for each of the keywords contained in the
   *  query.
   */
  private Ant[] getSearchAnts(RequestId rid, SearchRequest request, NestId requestor)
  {
    /* Compute hashed keywords */
    HashedKeyword keyhash[] = computeHashedKeywords(request.query);
    
    /* Create an ant for each of the keywords in the request */
    Ant[] ants = new Ant[keyhash.length];
    for (int i = 0; i < keyhash.length; i++) {
      ants[i] = new Gnutant(rid, request.query, keyhash[i], requestor, 
        explorationProbability, trailAttempts, trailDev, explorationAttempts,
        maxResources, maxProviders);
    }
    return ants;
  }
  
  /**
   *  Returns a set of insert ants, one for each of the keywords contained in the
   *  insert request.
   */
  private Ant[] getInsertAnts(RequestId rid, InsertRequest request, NestId requestor)
  {
    /* Compute hashed keywords */
    HashedKeyword keyhash[] = computeHashedKeywords(request.keywords);
    
    /* Create an ant for each of the keywords in the request */
    Ant[] ants = new Ant[keyhash.length];
    for (int i = 0; i < keyhash.length; i++) {
      // FIXME Remove comment to create insants; currently not available
      // ants[i] = new Insant(keyhash[i], request.urlResource);
    }
    return ants;
  }

  /**
   *  Compute keyword hashes from the specified string.
   *
   *  @param keywordString  a keyword string for which to compute hashes.
   *  @return  an array of <code>HashedKeyword</code> values corresponding to
   *    the specified keywords.
   */
  private HashedKeyword[] computeHashedKeywords(String keywordString)
  {
    /* Fragment the keywordString into keywords, separated by
     * spaces, underscore, dash and dot. */
    StringTokenizer tokenizer = new StringTokenizer(keywordString, TOKENS);
    HashedKeyword keyhash[] = new HashedKeyword[tokenizer.countTokens()];
    for (int i = 0; tokenizer.hasMoreTokens(); i++) {
      /* Obtain keyword hashes from the keywordString. */
      keyhash[i] = (HashedKeyword) factory.get(HashedKeyword.NAME);
      keyhash[i].init(tokenizer.nextToken());
    }
    return keyhash;
  }

} // END GnutantFactory

