/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package ants.gnutant;

import java.util.*;

import anthill.factory.Factory;
import anthill.nest.Nest;
import anthill.storage.*;
import antsim.*;

/**
 *  Classes implementing this interface are used to generate pre-defined
 *  resources at the beginning of a simulation.
 *  
 *  @author Alberto Montresor
 *  @version $Revision: 1.5 $
 */
public class GnutantResourceDistribution
implements ResourceDistribution
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Random generator */
  private static Random random = new Random();
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** 
  * If the Selection parameter assumes this value, resources 
  * are selected randomly (with uniform distribution) from the 
  * resource set.
  */
  public static final String UNIFORM = "Uniform";
  
  /** 
  * If the Selection parameter assumes this value, resources 
  * are selected randomly (with exponential distribution) from 
  * the resource set.
  */
  public static final String EXPONENTIAL = "Exponential";
  
  /** 
  * 
  */
  private static final int RANDOM_UNIFORM = 0;
  
  /** 
  * 
  */
  private static final int RANDOM_EXPONENTIAL = 1;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameter identifiers
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
  * String name of the parameter used to determine the inital number
  * of documents to be installed at each nest.
  */
  public static final String PAR_INITIALDOCS = "InitialDocumentsPerNest";
  
  /** 
  * String name of the parameter used to determine the type of
  * selection you can use. The possible values are: "Sequntial",
  * "Uniform", "Exponential"
  * 
  */
  public static final String PAR_SELECTION = "Selection";
  
  /** 
  * String name of the parameter used to determine vary the mean of
  * the keyword selection distribution.
  */
  public static final String PAR_MEAN = "Mean";
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** 
  * The inital number of documents to be installed at each nest.
  */
  private int initDocsPrNest; 
  
  /**
  * The kind of distribution to be used when selecting documents
  * to be placed in the system.
  */
  private int selection;
  
  /**  */
  private double mean;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Factory object used to read configuration parameters and classes */
  private Factory factory;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
  * 
  */
  public GnutantResourceDistribution(Factory factory, String name)
  {
    this.factory = factory;
    clear();
  }
  
  /**
  * Restores the object to its initial state.
  */
  public void clear()
  {
    /* Number of documents per nest */
    initDocsPrNest = factory.intValue(NAME + "." + PAR_INITIALDOCS);

    /* Determine the distribution used when distributing values */
    String sel = factory.stringValue(NAME + "." + PAR_SELECTION);
    if (sel.equals(UNIFORM))
      selection = RANDOM_UNIFORM;
    else if (sel.equals(EXPONENTIAL))
      selection = RANDOM_EXPONENTIAL;
    
    /* Mean value for the keyword selection distribution (geometric) */
    if (selection == RANDOM_EXPONENTIAL) {
      mean = factory.doubleValue(NAME + "." +PAR_MEAN);
    }
    
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
  * Distributes resources in a P2P network.
  */
  public void distributeResources(ResourceSet resourceSet, Peernet peernet)
  {
    List resources = resourceSet.getResources();
    List nests = peernet.getNests();

    if (selection == RANDOM_UNIFORM) {
    
      for (int i = 0; i < nests.size(); i++) {
        for (int j = 0; j < initDocsPrNest; j++) {
          
          /* Get the nest from the list*/
          Nest nest = (Nest) nests.get(i);
          
          /* Obtain a random keyword from the resource set */
          int index = random.nextInt(resources.size());
          String keyw = getKeyword(resources, index);
            
          /* Creates an actual resource */
          UrlResource resource = getUrlResource(nest, keyw);
          GenericStorage storage = (GenericStorage) nest.createResourceManager(Storage.DOC_STORAGE, null);
          try {
            storage.addResource(resource.getKey(), resource);
          } catch(CapacityException e) {
            System.err.println("Capacity exceeded in storage " + Storage.DOC_STORAGE + " located ad nest " + nest.getNestId());
            e.printStackTrace();
          } 
        }
      }

    } else {

      double beta = 1.0 - (1.0/mean);
      double logbeta = Math.log(beta);
      double powbeta = Math.pow(beta, resources.size() - 1);

      for (int i = 0; i < nests.size(); i++) {
        for (int j = 0; j < initDocsPrNest; j++) {

          /* Get the nest from the list*/
          Nest nest = (Nest) nests.get(i);
          
          /* Obtain a random keyword from the resource set */
          int index = compute(powbeta, logbeta);
          String keyw = getKeyword(resources, index);
          
          /* Creates an actual resource */
          UrlResource resource = getUrlResource(nest, keyw);
          GenericStorage storage = (GenericStorage) nest.createResourceManager(Storage.DOC_STORAGE, null);
          try {
            storage.addResource(resource.getKey(), resource);
          } catch(CapacityException e) {
            System.err.println("Capacity exceeded in storage " + Storage.DOC_STORAGE + " located ad nest " + nest.getNestId());
            e.printStackTrace();
          } 
        }
      }
      
    }
    
    
  }

    
  /**
  * Construct URL path string. 
  */
  private UrlResource getUrlResource(Nest nest, String keyw)
  {
    StringBuffer buf = new StringBuffer();
    buf.append("anthill://");
    buf.append(nest.getNestId());
    buf.append("/");
    buf.append(keyw);
    return new UrlResource(buf.toString());
  }
    
    
  /**
  * Returns the keyword stored in the given position.  If the given
  * index position is outside the table range, it will wrap to a legal
  * position using modulo table size.
  *
  * @param index  position into the keywordList.
  * @return a <code>String</code> containing a set of keywords
  *  associated with the given index position in the keywordList.
  */
  private String getKeyword(List resources, int index)
  {
    return (String) resources.get(index % resources.size());
  }
  
  /**
  * gnuplot:
  *  m=14  (Lowest legal value)
  *  b=1.0-(1.0/m)
  *  plot [x=0:1] log(x+b**(1+n)*(1-x))/log(b)
  */
  private int compute(double powbeta, double logbeta)
  {
    double x = random.nextDouble();
    double key = Math.log(x + powbeta * (1-x)) / logbeta;
    return (int) Math.round(key);
  }
  
  
} // END GnutantResourceDistribution

