/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package ants.gnutant;

import java.util.*;

import anthill.*;
import anthill.factory.Factory;
import anthill.nest.*;
import anthill.storage.*;
import anthill.util.Clearable;
import antsim.*;
import antsim.impl.ScenarioImpl;

/**
 *  Listens for responses from <code>ReplyGnutant</code> and log the
 *  statistics.  It also initiates a download phase of a certain
 *  percentage of the found documents.
 *
 *  @author Hein Meling
 *  @version $Revision: 1.7 $
 */
public class GnutantResponseMonitor
implements Clearable, ResponseMonitor, ResponseListener
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameter identifiers
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
  * Parameter name used to read the percentage of documents to
  * download upon receiving a set of documents references.
  */
  public static final String PAR_DOWNLOADPROB = "DownloadProbability";


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Percentage of the satisfied requests to download */
  private float dlProb;

  /** Number of rounds to be performed */
  private int rounds;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  private static Random random = new Random();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Stats */
  private Stats stats;

  /** Responses (requestId -> responseMap) */
  private HashMap responses;

  /** Time */
  private long time;

  /** Factory object used to read configuration parameters and classes */
  private Factory factory;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor / reinitialization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
  * Creates a new custom Gnutant response monitor.
  */
  public GnutantResponseMonitor(Factory factory, String name)
  {
    this.factory = factory;
    clear();
  }


  /**
  * Restores the object to its initial state.
  */
  public void clear() 
  {
    dlProb = factory.floatValue(NAME + "." + PAR_DOWNLOADPROB);
    rounds = factory.intValue(Scenario.NAME + "." + ScenarioImpl.PAR_REGULARROUNDS);

    if (responses == null) {
      responses = new HashMap();
    } else {
      responses.clear();
    }
    time = 0;
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ResponseMonitor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
  * 
  */
  public void setStatistics(Stats stats)
  {
    this.stats = stats;
    stats.add("TotalRequests", 0);
    stats.add("SatisfiedRequests", 0);
    stats.add("TotalResponses", 0);
    stats.add("LocalFound", 0);
    stats.add("TotalTime", 0);
  }

  /**
  * 
  */
  public void tick()
  {
    time++;
  }

  /**
  * 
  */
  public void tick(int v)
  {
    time += v;
  }

  /**
  * Must be invoked prior to invoking a request on a nest, to ensure
  * that the given request is stored to enable lookup on responses.
  */
  public void startRequest(RequestId rid)
  {
    if (responses.containsKey(rid))
      throw new IllegalArgumentException("Duplicated request: " + rid);
    responses.put(rid, new ResponseMap(time));
    stats.inc("TotalRequests");
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from ResponseListener
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
  * Upcall method used to notify this listener that the given request
  * has returned the specified set of resources.
  * 
  * @param rid the request identifier for the returned resources.
  * @param resources the resources found.
  */
  public void returnResources(RequestId rid, Resource[] resources) 
  {
    /*
    * The timeout should be less than the rounds because after 'rounds'
    * iterations we may not receive more responses for the specified 'rid'.
    */
    try {
      int timeout = rounds / 8;
//       Log.println("timeout: "+ timeout);
      if (returnResources(rid, resources, timeout)) {
        if (dlProb > 0.0)
          downloadResources(rid);
        /* Remove old responses to avoid memory problems */
        ResponseMap map = (ResponseMap) responses.remove(rid);
        map.responses.clear();
      }
    } catch(CapacityException ce) {
      ce.printStackTrace();
    }
  }


  /**
  * Upcall method used to notify this listener that the given request
  * has finished to return resources.
  */
  public void complete()
  {
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Private methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  private void downloadResources(RequestId rid) throws CapacityException
  {
    Nest nest = ((Rid) rid).getNest();
    GenericStorage docStore = (GenericStorage) nest.createResourceManager(Storage.DOC_STORAGE, null);
    GenericStorage urlStore = (GenericStorage) nest.createResourceManager(Storage.URL_STORAGE, null);
    
    for (Iterator respIter = responses(rid); respIter.hasNext(); ) {
      /* Get the resource corresponding to this response */
      UrlResource resource = (UrlResource) respIter.next();

      /* Don't download resources that already exist locally */
      DocumentKey docKey = (DocumentKey) resource.getKey();
      if (docStore.containsResource(docKey) || urlStore.containsResource(docKey))
        continue;

      /* Apply download probability according to a uniform distribution */
      if (random.nextFloat() < dlProb) {
        /* Download the file (resource) corresponding to this response */

        /* Insert the downloaded document in the local document storage. */
        UrlResource localUrl = localizeUrlResource(nest, resource);
        try {
          docStore.addResource(localUrl.getKey(), localUrl);
        } catch(CapacityException e) {
          System.err.println("Capacity exceeded in storage " + Storage.DOC_STORAGE + " located ad nest " + nest.getNestId());
          e.printStackTrace();
        } 
        
        /* Notify other nests that I have downloaded a document. */
        nest.request(null, getInsantRequest(nest, localUrl.getFile()), null);
      }
    }
  }


  /**
  * Construct URL path string. 
  */
  private UrlResource localizeUrlResource(Nest nest, UrlResource urlResource)
  {
    StringBuffer buf = new StringBuffer();
    buf.append("local://");
    buf.append(nest.getNestId());
    buf.append("/");
    buf.append(urlResource.getFile());
//     Log.println("\n%%% Copied file: " + buf.toString());
    return new UrlResource(buf.toString(), (DocumentKey) urlResource.getKey());
  }


  /**
  * Construct an InsertRequest.
  */
  private InsantRequest getInsantRequest(Nest nest, String fileName)
  {
    StringBuffer buf = new StringBuffer();
    buf.append("insant://");
    buf.append(nest.getNestId());
    buf.append("/");
    buf.append(fileName);
    return new InsantRequest(buf.toString(), fileName);
  }


  /**
  * Upcall method used to notify this listener that the given request
  * has returned the specified set of resources.
  *
  * @param rid the request identifier for the returned resources.
  * @param resources the resources found.
  * @param timeout the timeout value
  * @return true  if the timeout has passed.
  */
  private boolean returnResources(RequestId rid, Resource[] resources, int timeout)
  {
    ResponseMap map = (ResponseMap) responses.get(rid);
    if (map != null) {
      long diff = time - map.time;
      if (map.isEmpty()) {
        stats.inc("SatisfiedRequests");
        if (diff == 0)
          stats.inc("LocalFound");
        stats.add("TotalTime", diff);
      }
      for (int i = 0; i < resources.length; i++) {
        UrlResource urlResource = (UrlResource) resources[i];
        if (map.responses.put(urlResource.getKey(), urlResource) == null) {
          /*
          * The resource was not present in the set of received
          * responses; increment number of unique resources found.
          */
          stats.inc("TotalResponses");
        }
      }
      if (diff >= timeout) 
        return true;
    }
    return false;
  }


  /**
  * Returns the responses associated with the specified response
  * identifier.
  *
  * @param rid the request identifier whose responses to return.
  * @return an iterator over the responses associated with the
  *   given response identifier.
  */
  private Iterator responses(RequestId rid)
  {
    ResponseMap map = (ResponseMap) responses.get(rid);
    if (map == null)
      throw new IllegalArgumentException("Unknown request: " + rid);
    return map.responses.values().iterator();
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Inner class 
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
  * Response collection
  */
  private class ResponseMap
  {
    /** Request generation time */
    long time;

    /** Map of unique responses (dockey -> resource) */
    Map responses;

    ResponseMap(long time)
    {
      this.time = time;
      responses = new HashMap();
    }

    boolean isEmpty()
    {
      return responses.isEmpty();
    }

  } // END ResponseMap

} // END GnutantResponseMonitor
