/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package ants.jxtagnutant;

import java.io.*;
import java.util.*;

import anthill.*;
import anthill.storage.*;
import anthill.types.HashedKeyword;

/**
 *
 *  @author Hein Meling
 *  @author Francesco Russo (frusso@cs.unibo.it)
 *  @version $Revision: 1.13 $
 */
public class SearchGnutant
implements Ant, Serializable
{
  private static Random rand = new Random();


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
  * Max number of times the ant will try and move at random when nests listed in the
  * pheromone data structure are not reacheable
  */
  private static final int MAXTRIES = 5;

  /** Max number of times to backtrack before giving up. */
  private static final int MAXRESULTS = 4;

  /** Max number of times to backtrack before giving up. */
  private static final int BACKTRACK_LIMIT = 10;

  /** Backtracking on or off. */
  private static final boolean backtracking = true;

  ////////////////////////////////////////////////////////////////////////////////////////////
  // State fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Variable path contains the list of visited nodes, allowing the ant
   *  to avoid visiting the same nests multiple times and backtrack to
   *  the originating nest.
   */
  private ArrayList path;

  /** The identifier of the request */
  private RequestId rid;

  /** The query associated with this ant */
  private Object query;

  /** The hash value of the keyword searched by this ant */
  private HashedKeyword keyhash;

  /** Resource to be returned */
  private HashSet resources;

  /** Number of backtracks performed so far by this ant. */
  private int backtracks;

  long rsize;
  int tries;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Instantiate a new SearchGnutant by setting the query to be satisfied and the
   *  hash value to be searched.
   */
  public SearchGnutant(RequestId rid, Object query, HashedKeyword keyhash)
  {
  	/* Store initialization parameters */
  	this.rid = rid;
  	this.query = query;
  	this.keyhash = keyhash;
  	backtracks = 0;
  	path = new ArrayList();
  	resources = new HashSet();
	tries = 0;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  // Documentation inherited from interface
  public String getSpecies()
  {
  	return "jxtagnutant";
  }

  
  /**
   *
   */
  public void init(AntView view)
  {
  }
  
  /**
   *  This method encapsulate the computation performed by the ant.
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
  	path.add(view.getNestId());

  	/*
   	*  Obtains the local pheromone contained in this nest; if not found,
   	*  creates a new pheromone object and installs it locally
   	*/
  	System.out.println("--- SearchGnutant: run() - getting pheromone object"); 
  	//GnutantPheromone pher = (GnutantPheromone) view.getPheromone();
	GnutantPheromone pher = (GnutantPheromone) view.getResourceManager( Storage.PHR_STORAGE );
  	System.out.println("--- SearchGnutant: run() - done!");
  	
  	if (pher == null) 
  	{
	    //pher = new GnutantPheromone(view);
	    pher = (GnutantPheromone) view.createResourceManager( Storage.PHR_STORAGE );
		pher.init( view );
	    System.out.println("--- SearchGnutant: run() - set new pheromone object!");
  	} 
	
  	/* Perform a local check of the storages */
  	System.out.println("--- SearchGnutant: run() - about to check the Storage");
  	checkStorage(view);
  	System.out.println("--- SearchGnutant: run() - done!");
	
  	if (rsize > MAXRESULTS) 
  	{
	    Resource[] res = new Resource[resources.size()];
	    resources.toArray(res);
	    view.addAnt(new BackwardGnutant(keyhash, res, (ArrayList) path.clone()));
	    System.out.println("--- SearchGnutant: run() - BackwardAnt added to the Nest");
  	}  
  	
  	/* 
   	*  Get the a set of potential next nests --
   	*  based on the existence of the given keyhash
   	*  or according to the closeness relation to the keyhash 
   	*/
  	ArrayList nxtList = pher.getNextNest(keyhash, path);
  	
  	for (int i = 0; i < nxtList.size(); i++) 
  	{
		try 
	  	{
	  		//jxtaimpl.ids.NestId dstId = (jxtaimpl.ids.NestId) nxtList.get(i);
	  		//jxtaimpl.ids.NestId srcId = (jxtaimpl.ids.NestId) ((jxtaimpl.ids.RequestId)rid).getNestId();
	  		//if ( ! (srcId.toString().trim().toLowerCase().equals(dstId.toString().trim().toLowerCase())) )
	  		
			NestId srcId = rid.getNestId(); 
			NestId dstId = (NestId)nxtList.get(i); 
			if( !srcId.equals(dstId) )
			{
				System.out.println("--- SearchGnutant: run() - moving to another Nest");
				view.move((NestId) nxtList.get(i));
				System.out.println("--- SearchGnutant: run() - done!");
				/* Successful move to new nest */
				return;
	  		}
		} 
	  	catch (IOException e) 
		{ 
	  		/* Try move to an alternative nest */
	  		continue;
		}
  	} // end of for
	
  	System.out.println("--- SearchGnutant: run() - no useful neighbours, RANDOM MOVE");
  	move( view );
  }

  /**
   *  
   */
  private void checkStorage(AntView view)
  {
  	QueryStorage docStorage = null;
  	QueryStorage urlStorage = null;
  	ResourceManager resource = null;
	
  	try
  	{
	    //docStorage = (QueryStorage) view.getStorage(Storage.DOC_STORAGE);
	    //urlStorage = (QueryStorage) view.getStorage(Storage.URL_STORAGE);
		docStorage = (QueryStorage) view.getResourceManager( Storage.DOC_STORAGE );
		resource = (ResourceManager) view.getResourceManager( Storage.URL_STORAGE );
  	}
  	catch( StorageNotFoundException snfe )
  	{
	    System.out.println("--- SearchGnutant: checkStorage() - StorageNotFoundException!!!");
  	}
	catch( ResourceManagerException rme )
  	{
	    System.out.println("--- SearchGnutant: checkStorage() - ResourceManagerdException!!!");
  	}
  	
  	/* Create an empty array list for the new resources */
  	HashSet newResources = new HashSet();
  	
  	System.out.println("--- SearchGnutant: checkStorage() - created HashSet");
  	
  	/* Add resources found in the doc storage */
  	if (docStorage != null) 
  	{
	    System.out.println("--- SearchGnutant: checkStorage() - quering DOC_STORAGE");  
	    Resource[] docs = docStorage.getResources(query);
	    if (docs != null) 
	    {
		    for (int i=0; i < docs.length; i++) 
			{
	  			if (resources.add(docs[i])) 
	  			{
					System.out.println("--- SearchGnutant: checkStorage() - adding a DOC_resource");
	  				newResources.add(docs[i]);
	  				rsize = rsize + docs[i].size();
	  			}
			}
		}
  	}
  	
  	/* Add resources found in the url storage */
  	if (urlStorage != null) 
  	{
	    urlStorage = ((QueryStorage)resource);
	    System.out.println("--- SearchGnutant: checkStorage() - quering URL_STORAGE");    
	    Resource[] urls = urlStorage.getResources(query);
	    if (urls != null) 
	    {
		    for (int i=0; i < urls.length; i++) 
			{
	  			if (resources.add(urls[i])) 
	  			{
					System.out.println("--- SearchGnutant: checkStorage() - adding an URL_resource");
	  				newResources.add(urls[i]);
	  				rsize = rsize + urls[i].size();
	  			}
			}
		}
  	}
	
  	/*  
   	*  If there are new resources, send a ReplyGnutant.
   	*/
  	if (newResources.size() > 0) 
  	{
	    System.out.println("--- SearchGnutant: checkStorage() - about to send a ReplyAnt"); 
	    Resource[] news = new Resource[newResources.size()];
	    newResources.toArray(news);
	
	    /*
	    *  Report the found resources directly to the originating nest.
		*/
		ReplyGnutant rgant = new ReplyGnutant(rid, news, (NestId) path.get(0));
		System.out.println("--- SearchGnutant: checkStorage() - ReplayAnt created!");
		view.addAnt(rgant);
		System.out.println("--- SearchGnutant: checkStorage() - ReplayAnt added to the Nest");
  	}
	
  	/*
   	*  Once the TTL have been exhausted we do a backward walk of
   	*  the path in order to update the distributed index.
   	*/
	if (view.getTTL() == 0) 
  	{
	    Resource[] res = new Resource[resources.size()];
	    resources.toArray(res);
	    /*
	    *  BackwardGnutant is used to update the distributed index.
	    *  Since we have reached the TTL we don't need to clone the
	    *  path even though BackwardGnutant removes path entries as
	    *  it backtracks.
	    */
	    view.addAnt(new BackwardGnutant(keyhash, res, (ArrayList) path.clone()));
	    System.out.println("--- SearchGnutant: checkStorage() - BackwardAnt added to the Nest");
  	}
  }
	

  /**
  * This private method is used when the ant doesn't know where to go may-be
  * due to a lack of information stored into the current nest pheromone object.
  * What it does is simply try to reach this nest's known neighbours. 
  * Warning: this method could lead an ant waiting indefinitely since the set
  * of neighbours could even be or became empty in the meantime! 
  */
  private void move(AntView view)
  {
  	if(tries < MAXTRIES) 
	{
		tries++;
  		Neighbour[] neighbours = view.getNeighbours();
   	
	  	while (neighbours.length < 1)
	   	{
	    	try
	      	{
		        Thread.sleep(4000);
		    }
	      	catch(InterruptedException ie)
	      	{
		        ie.printStackTrace();
		    }
	      	neighbours = view.getNeighbours();
	    }
	    int next = rand.nextInt(neighbours.length);
	    try 
	    {
			//jxtaimpl.ids.NestId dstId = (jxtaimpl.ids.NestId)neighbours[next].getNestId();
			//jxtaimpl.ids.NestId srcId = (jxtaimpl.ids.NestId) ((jxtaimpl.ids.RequestId)rid).getNestId();
			//if ( ! (srcId.toString().trim().toLowerCase().equals(dstId.toString().trim().toLowerCase())) )
			
			NestId srcId = neighbours[next].getNestId();
			NestId dstId = rid.getNestId();
			if( !srcId.equals(dstId) )
	  		{
	    		System.out.println("--- SearchGnutant: move() - goForward - moving on, nothing has been found!");
	      		view.move(neighbours[next].getNestId());
			}
			else
			{
				System.out.println("--- SearchGnutant: move() - found my home nest, looking for a new one!");
	      		move( view );
			}
	    } 
	    catch (IOException e) 
	    {
	      	System.out.println("--- SearchGnutant: move() - nest unreachable, looking for a new one!");
	      	move( view );
	    }   
  	}
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Externalization methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void writeExternal(ObjectOutput out)
  throws IOException
  {
  // FIXME To be implemented
  }
  
  public void readExternal(ObjectInput in)
  {
  // FIXME To be implemented
  }
  
} // END SearchGnutant


