/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package ants.messor;

import java.io.*;
import java.util.*;

import anthill.*;
import anthill.util.Log;


/**
 *
 *  @author Alberto Montresor
 *  @version $Revision: 1.7 $
 */
public class MessorAnt
implements Ant, Serializable
{

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constant representing the SearchMax status
   */
  private static final int ST_SEARCHMAX = 0;
  
  /**
   *  Constant representing the SearchMin status
   */
  private static final int ST_SEARCHMIN = 1;
  
  /**
   *  Constants representing the MoveMax status
   */
  private static final int ST_MOVEMAX = 2;
  
  /**
   *  Constants representing the MoveMin status
   */
  private static final int ST_MOVEMIN = 3;
  
  /**
   *  Constants representing the MoveMin status
   */
  private static final int ST_LEAVETASKS = 4;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * Random number generator.
   */ 
  private static Random random = new Random();
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  When in MaxSearching status, probability to explore the network 
   *  instead to go towards nodes with max load.
   */
  private float maxExploringProb;
  
  /** 
   *  When in MinSearching status, probability to explore the network 
   *  instead to go towards nodes with min load.
   */
  private float minExploringProb;
  
  /** 
   *  When in MaxSearching status, standard deviation for selecting
   *  the next node using a normal distribution.
   */
  private double maxSearchDev;
  
  /** 
   *  When in MinSearching status, standard deviation for selecting
   *  the next node using a normal distribution.
   */
  private double minSearchDev;

  
  /**
   *  When in MaxSearching status, probability to keep searching when
   *  a node with a load greater than the given treshold has been 
   *  reached. 
   */
  private float maxKeepSearchingProb;
  
  /**
   *  When in MaxSearching status, probability to keep searching when
   *  a node with a load greater than the given treshold has been 
   *  reached. 
   */
  private float minKeepSearchingProb;
  
  /**
   *  Target ratio between the load of overloaded and underloaded 
   *  nests when in SearchMax status. When an ant is looking for 
   *  overloaded machines, it keeps moving until the ratio between 
   *  the load observed in the nest with maximum load and the nest 
   *  with minimum load do not reach this target value.
   */
  private float maxTargetRatio;
  
  /**
   *  Target ratio between the load of overloaded and underloaded 
   *  nests when in SearchMin status. When an ant is looking for 
   *  underloaded machines, it keeps moving until the ratio between 
   *  the load observed in the nest with maximum load and the nest 
   *  with minimum load do not reach this target value.
   */
  private float minTargetRatio;
  
  /**
   *  Percentage of tasks to be move from one site to another.  
   */
  private float transportPercentage;
  
  /**
   *  Max number of tasks that can be transported by one node to another  
   */
  private int maxTransport;
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Status
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   * Status of the ant; can get the values SearchMax, SearchMin
   *  and Transport
   */
  int status;

  /**
   *  Last nests visited; this array is organized as a circular queue.  
   */
  NestId[] visited;
  
  /**
   *  Load associated to the last visited nests. Each position in this
   *  array corresponds to the same position in the <CODE>visited</CODE> 
   *  array.
   */
  int[] loads;

  /** 
   *  Index of the next position where to write nests.
   */
  int index;
  
  /**
   *  Contains the maximum load among those recently visited by
   *  this ant.
   */
  int maxLoad;
  
  /**
   *  Contains the identifier of the nest in which the maximum
   *  load has been observed.
   */
  NestId maxNest;

  /**
   *  Contains the minimum load among those recently visited by
   *  this ant.
   */
  int minLoad;

  /**
   *  Contains the identifier of the nest in which the minimum
   *  load has been observed.
   */
  NestId minNest;
  
  
  /**
   *  Tasks to be transported from one nest to another
   */
  List tasks;
  
  /**
   *  Ant identifier; currently used only for debug.
   */
  AntId aid;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   * 
   */
  public MessorAnt(
    float maxExploringProb, float minExploringProb,
    double maxSearchDev, double minSearchDev,
    float maxKeepSearchingProb, float minKeepSearchingProb,
    float maxTargetRatio, float minTargetRatio,
    float transportPercentage, int maxTransport, int size)
  {
    /* Stores parameters */
    this.maxExploringProb = maxExploringProb;
    this.minExploringProb = minExploringProb;
    this.maxSearchDev = maxSearchDev;
    this.minSearchDev = minSearchDev;
    this.maxKeepSearchingProb = maxKeepSearchingProb;
    this.minKeepSearchingProb = minKeepSearchingProb;
    this.maxTargetRatio = maxTargetRatio;
    this.minTargetRatio = minTargetRatio;
    this.transportPercentage = transportPercentage;
    this.maxTransport = maxTransport;
    
    /* Creates visited array and load array */
    visited = new NestId[size];
    loads = new int[size];
    index = 0;
    
    /* Initially, the task list is empty */
    status = ST_SEARCHMAX;
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Ant
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Returns the identifier of the "species" of this ant.
   *  Different ant species may have access to different
   *  pheromone information in nests.
   */
  public String getSpecies()
  {
    return "MessorAnt";
  }

  public void init(AntView view)
  {
    /* Obtain storage classes */
    TaskStorage storage = 
      (TaskStorage) view.createResourceManager(TaskStorage.NAME);

    /* Set the max load and min load to fake values */
    maxLoad = minLoad = storage.getLoad();
    maxNest = minNest = view.getNestId();
 
    aid = view.createAntId();   
  }

  /**
   *  This method encapsulate the computation performed by the
   *  ant. 
   *
   *  @param view reference to the AntView object which
   *    constitutes the controlled environment of the ant
   */
  public void run(AntView view)
  {
    /* Obtain storage classes */
    TaskStorage storage = 
      (TaskStorage) view.createResourceManager(TaskStorage.NAME);
    MessorPheromone pheromone = 
      (MessorPheromone) view.createResourceManager(MessorPheromone.NAME);
    pheromone.update(view.getNestId(), view.getNeighbours());

    /* Debug */
    Log.println(500, aid + " at " + view.getNestId() + ": load= " + storage.getLoad()); 
    
    /* The ant commits suicide if the nest is overloaded */
    if (status != ST_LEAVETASKS && pheromone.isOverladed())
      return;
    
    switch (status) {
    case ST_SEARCHMAX:
      doSearchMax(pheromone, storage, view);
      break;
    case ST_SEARCHMIN:
      doSearchMin(pheromone, storage, view);
      break;
    case ST_MOVEMAX:
      doMoveMax(pheromone, storage, view);
      break;
    case ST_MOVEMIN:
      doMoveMin(pheromone, storage, view);
      break;
    case ST_LEAVETASKS:
      doLeaveTask(pheromone, storage, view);
      break;
    }
    
    /* Updates pheromone information */
    pheromone.update(view.getNestId(), visited, loads);
    
    /* Updates visited nodes */
    visited[index] = view.getNestId();
    loads[index] = storage.getLoad();
    index = (index + 1) % visited.length;
    
  }

  /**
   *  Algorithm to be followed when the ant is in the 
   *  ST_SEARCHMAX status.
   */
  private void doSearchMax(MessorPheromone pheromone, TaskStorage storage, 
    AntView view)
  {
    Log.println(500, aid + " at " + view.getNestId() + ": doSearchMax load= " + 
      storage.getLoad() + ", maxLoad= " + maxLoad + ", minLoad= " + minLoad +
      ", maxNest= " + maxNest + ", minNest= " + minNest); 
    int localLoad = storage.getLoad();
    if (localLoad > maxLoad) {
      /*
       *  This nest has the maximum load found so far; we 
       *  register this value in maxLoad and maxNest.
       */
      maxLoad = localLoad;
      maxNest = view.getNestId();
    }
    if (localLoad < minLoad) {
      /*
       *  This nest has the minimum load found so far; we 
       *  register this value in maxLoad and maxNest.
       */
      minLoad = localLoad;
      minNest = view.getNestId();
    }
    
    if ( ( (float) (maxLoad - minLoad) / maxLoad ) >= maxTargetRatio ) {
      /*
       *  The ant has found a sufficient ratio between the maximum
       *  nest and the minimum nest; the ant may continue searching
       *  depending on maxKeepSearching
       */
      if (!tossCoin(maxKeepSearchingProb)) {
        /* 
         *  The ant stop searching overloaded nests, and 
         *  start searching underloaded nests.
         */
        status = ST_SEARCHMIN;
        //minLoad = maxLoad;
        //minNest = maxNest;
        doSearchMin(pheromone, storage, view);
        return;
      }
    }
    
    /* Selects between exploring behavior and searching behavior */
    boolean exploring = tossCoin(maxExploringProb);
    
    /* Selects a nest */
    NestId next;
    boolean moved = false;
    while (!moved) {
      if (exploring) {
        /* We are exploring; just select one node at random */
        next = pheromone.getRandomNest();
        Log.println(500, aid + " at " + view.getNestId() + ": Moving random, " + next);
      } else {
        next = pheromone.getNextMaxNest(maxSearchDev);
        Log.println(500, aid + " at " + view.getNestId() + ": Moving to max, " + next);
      }
      if (next == view.getNestId())
        continue;
      try {
        view.move(next);
        moved = true;
        Log.println(500, aid + " at " + view.getNestId() + ": Moved to " + next); 
      } catch (IOException e) { }
    }
    
  }
  
  /**
   *  Algorithm to be followed when the ant is in the 
   *  ST_SEARCHMIN status.
   */
  private void doSearchMin(MessorPheromone pheromone, TaskStorage storage, 
    AntView view)
  {
    Log.println(500, aid + " at " + view.getNestId() + ": doSearchMin load= " + 
      storage.getLoad() + ", maxLoad= " + maxLoad + ", minLoad= " + minLoad +
      ", maxNest= " + maxNest + ", minNest= " + minNest); 
    int localLoad = storage.getLoad();
    if (localLoad < minLoad) {
      /*
       *  This nest has the minimum load found so far; we 
       *  register this value in maxLoad and maxNest.
       */
      minLoad = localLoad;
      minNest = view.getNestId();
    }
    if (localLoad > maxLoad) {
      /*
       *  This nest has the maximum load found so far; we 
       *  register this value in maxLoad and maxNest.
       */
      maxLoad = localLoad;
      maxNest = view.getNestId();
    }

    if ( ( (float) (maxLoad - minLoad) / maxLoad ) >= maxTargetRatio ) {
      /*
       *  The ant has found a sufficient ratio between the maximum
       *  nest and the minimum nest; the ant may continue searching
       *  depending on minKeepSearching
       */
      if (!tossCoin(minKeepSearchingProb)) {
        /* 
         *  The ant stop searching overloaded nests, and 
         *  start searching underloaded nests.
         */
        status = ST_MOVEMAX;
        doMoveMax(pheromone, storage, view);
        return;
      }
    }
    
    /* Selects between exploring behavior and searching behavior */
    boolean exploring = tossCoin(maxExploringProb);
    
    /* Selects a nest */
    NestId next;
    boolean moved = false;
    while (!moved) {
      if (exploring) {
        /* We are exploring; just select one node at random */
        next = pheromone.getRandomNest();
        Log.println(500, aid + " at " + view.getNestId() + ": Moving random, " + next);
      } else {
        next = pheromone.getNextMinNest(minSearchDev);
        Log.println(500, aid + " at " + view.getNestId() + ": Moving to min, " + next);
      }
      if (next == view.getNestId())
        continue;
      try {
        view.move(next);
        moved = true;
        Log.println(500, aid + " at " + view.getNestId() + ": Moved to " + next); 
      } catch (IOException e) { }
    }
    
  }
  
  /**
   *  Algorithm to be followed when the ant is in the 
   *  ST_MOVEMAX status.
   */
  private void doMoveMax(MessorPheromone pheromone, TaskStorage storage, 
    AntView view)
  {
    Log.println(500, aid + " at " + view.getNestId() + ": doMoveMax load= " + 
      storage.getLoad() + ", maxLoad= " + maxLoad + ", minLoad= " + minLoad +
      ", maxNest= " + maxNest + ", minNest= " + minNest); 
    try {
      status = ST_MOVEMIN;
      view.move(maxNest);
      Log.println(500, aid + " at " + view.getNestId() + ": Moved to " + maxNest); 
    } catch (IOException e) {
      e.printStackTrace();
      /* FIXME: What happens if the max node is crashed in
       * between? We need to select other maximum nodes 
       * between those registered in the ant.
       */
    }
  }
  
  /**
   *  Algorithm to be followed when the ant is in the 
   *  ST_MOVEMMIN status.
   */
  private void doMoveMin(MessorPheromone pheromone, TaskStorage storage, 
    AntView view)
  {
    Log.println(500, aid + " at " + view.getNestId() + ": doMoveMin load= " + 
      storage.getLoad() + ", maxLoad= " + maxLoad + ", minLoad= " + minLoad +
      ", maxNest= " + maxNest + ", minNest= " + minNest); 
    /* 
     *  Extracts a number of tasks from the local storage,
     *  depending on the load difference between the max nest
     *  and the min nest and on the percentage of this difference
     *  that must be transported from a node to another.
     */
    int size = (int) ((storage.getLoad() - minLoad) * transportPercentage);
    if (size < 0) {
      size = 1;
    } else if (size > maxTransport) {
      size = maxTransport;
    }
    Log.println(500, aid + " at " + view.getNestId() + ": picking " +  size + " tasks"); 
    tasks = storage.getTasks(size);
    
    /* Try to move to min nest */
    try {
      status = ST_LEAVETASKS;
      view.move(minNest);
      Log.println(500, aid + " at " + view.getNestId() + ": Moved to " + minNest); 
    } catch (IOException e) {
      /* FIXME: What happens if the max node is crashed in
       * between? We need to select other maximum nodes 
       * between those registered in the ant.
       */
    }
  }
  
  /**
   *  Algorithm to be followed when the ant is in the 
   *  ST_LEAVETASK status.
   */
  private void doLeaveTask(MessorPheromone pheromone, TaskStorage storage, 
    AntView view)
  {
    Log.println(500, aid + " at " + view.getNestId() + ": doLeaveTasks load= " + 
      storage.getLoad() + ", tasks=" + tasks.size() + ", maxLoad= " + maxLoad + ", minLoad= " + minLoad +
      ", maxNest= " + maxNest + ", minNest= " + minNest); 
    storage.addTasks(tasks);
    tasks = null;
    maxLoad = storage.getLoad();
    minLoad = storage.getLoad();
    minNest = view.getNestId();
    maxNest = view.getNestId();
    status = ST_SEARCHMAX;
    doSearchMax(pheromone, storage, view);
  }
    
  /**
   *  Toss a coin returning true or false, with the specified
   *  probability of returning true.
   * 
   *  @param prob The probability of returning true
   */
  private boolean tossCoin(float prob)
  {
    float coin = random.nextFloat();
    return (coin < prob);
  }
  
} // END MessorAnt

