/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package antsim.impl;

import java.util.*;

import anthill.Resource;
import anthill.factory.Factory;
import anthill.nest.Nest;
import antsim.*;

/**
 *  This class implements a generic request generator. To create a 
 *  request, the generator selects randomly (with uniform probability) 
 *  a resource from a resource set, creates a generic request for that 
 *  resource and assign it randomly to one of the nests included in the 
 *  peernet. The number of requests that are generated for each invocation
 *  of method <CODE>generateRequests</CODE> can be configured through
 *  the <CODE>NumberOfRequests</CODE> parameter.
 * 
 *  @author Alberto Montresor
 *  @version $Revision: 1.6 $
 */
public class GenericRequestGenerator
implements RequestGenerator
{
 
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Random generator */
  private static Random random = new Random();

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Parameter name used to read the number of requests to be generated */
  public static final String PAR_NUMBERREQUESTS = "NumberOfRequests";
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  Number of requests to be generated whenever method
   *  <CODE>generateRequests</CODE> is invoked.
   */
  private int nrequests;  
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Factory object used to read configuration parameters and classes */
  private Factory factory;

  /** List of nests on which to generate requests */
  private List nests;

  /** Set of resources from which to select a request to be generated */
  private List resources;

  /** The response monitor object for this request generator */
  private ResponseMonitor monitor;


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Constructs a new generic request generator and reads parameter
   *  <CODE>NumberOfRequests</CODE> from the configuration files.
   *  
   *  @param factory the factory to be used to read configuration arguments
   *    and to create other configurable objects
   *  @param name the name that is associated to this instance in the
   *    configuration file
   */
  public GenericRequestGenerator(Factory factory, String name)
  {
    this.factory = factory;
    clear();
  }
  
  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    nrequests = factory.intValue(NAME + "." +PAR_NUMBERREQUESTS);
  }


  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from the RequestGenerator interface
  ////////////////////////////////////////////////////////////////////////////////////////////

  public void init(ResourceSet resourceSet, Peernet peernet, ResponseMonitor monitor)
  {
    nests = peernet.getNests();
    resources = resourceSet.getResources();
    this.monitor = monitor;
  }


  /**
   *  Generates a set of requests in the specified P2P network.
   *  Different implementations may adopt different generation and
   *  distribution patterns.
   */
  public void generateRequests()
  {
    /* Choose random nest and make a request */
    for (int j=0; j < nrequests; j++) {
      int nestIndex = random.nextInt(nests.size());
      int resourceIndex = random.nextInt(resources.size());
      Resource resource = (Resource) resources.get(resourceIndex);
      Object request = resource.getName();
      Nest nest = (Nest) nests.get(nestIndex);
      Rid rid = new Rid(nest);
      monitor.startRequest(rid);
      nest.request(rid, request, monitor);
    }
  }

  
} // END RequestGenerator

