/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package antsim.impl;

import java.util.*;

import anthill.factory.Factory;
import anthill.util.Log;
import antsim.*;

/**
 *  This is a basic implementation of the Peernet interface. A <CODE>PeernetImpl</CODE>
 *  is characterized by a certain size (i.e., the number of nests composing
 *  it) and a degree (i.e., the number of links connecting it to other
 *  nests). 
 *  <p>
 *  This <CODE>Peernet</CODE> implementation uses the <CODE>Factory</CODE> class to 
 *  dynamically instantiate objects implementing Nest; the <CODE>Factory</CODE> 
 *  instance must be opportunely instructed with an XML definition for <CODE>Nest</CODE>.
 *  
 *  @author Alberto Montresor
 *  @version $Revision: 1.5 $
 */
public class PeernetImpl
implements Peernet
{
  

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constants
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** 
   *  String name of the parameter used to determine the number of nodes included
   *  in this peernet.
   */
  public static final String PAR_SIZE = "Size";

  /** 
   *  String name of the parameter used to determine the degree of the nodes included
   *  in this peernet.
   */
  public static final String PAR_DEGREE = "Degree";

  /** 
   *  String name of the parameter used to determine the probability for
   *  a node to leave the peer network.
   */
  public static final String PAR_LEAVEPROB = "LeaveProbability";
  
  /** 
   *  String name of the parameter used to determine the probability for
   *  a node to join the peer network.
   */
  public static final String PAR_JOINPROB = "JoinProbability";
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random number generator */
  static Random random = new Random();
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** The size of the network, i.e. the number of nests included in it */
  private int size;

  /** The degree of the nests, i.e. the number of links to other nests */
  private int degree;

  /** 
   *  The join probability, i.e. the probability that a new node joins the 
   *  peer network at each round.
   */
  private float joinProb;
  
  /** 
   *  The leave probability, i.e. the probability that a node leaves the 
   *  peer network at each round.
   */
  private float leaveProb;

 
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /** Factory object used to read configuration parameters and classes */
  private Factory factory;
  
  /** Nest network */
  private List nests;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  Initializes a new simulated peer networks. The initial size and
   *  degree are read from the initialization parameters.
   *  
   *  @param factory the factory to be used to read configuration arguments
   *    and to create other configurable objects
   *  @param name the name that is associated to this instance in the
   *    configuration file
   */
  public PeernetImpl(Factory factory, String name)
  {
    this.factory = factory;
    clear();
  }
  

  /**
   * Restores the object to its initial state. 
   * New nests are instantiated when this method is called for the first
   * time; if the initialization arguments for the Nest interface have
   * been changed, the old nests are discarded and the new nests are 
   * created. In all other cases, the old nodes are cleared.
   */
  public void clear()
  {
    /*
     *  Read parameters.
     */
    size = factory.intValue(NAME + "." + PAR_SIZE);
    degree = factory.intValue(NAME + "." + PAR_DEGREE);
    leaveProb = factory.floatValue(NAME + "." + PAR_LEAVEPROB);
    joinProb = factory.floatValue(NAME + "." + PAR_JOINPROB);

    if (nests == null) {

      /*
       *  If the nest network is empty, updates the current nest configuration 
       *  version and creates a new nest array representing the network 
       */
      nests = new ArrayList(size);
      for (int i=0; i < size; i++) {
        nests.add(factory.get(SimulatedNest.NAME));
      }
      
    } else if (nests.size() < size) {

      /* 
       *  The new size is greater than the previous size; 
       *  clear previous one, create new nests
       */
      for (int i=0; i < nests.size(); i++) {
        SimulatedNest nest = (SimulatedNest) nests.get(i);
        nest.clear();
      }
      for (int i=nests.size(); i < size; i++)
        nests.add(factory.get(SimulatedNest.NAME));
    
    } else if (nests.size() > size) {
      
      /* 
       *  The new size is less than the previous size; 
       *  discard additional nests, clear remaining ones
       */
      while (nests.size() > size) {
        nests.remove(nests.get(nests.size()-1));
      }
      for (int i=0; i < nests.size(); i++) {
        SimulatedNest nest = (SimulatedNest) nests.get(i);
        nest.clear();
      }
      
    } else {

      /* 
       *  The new size is equal to the previous size; 
       *  just clear existing nests
       */
      for (int i=0; i < nests.size(); i++) {
        SimulatedNest nest = (SimulatedNest) nests.get(i);
        nest.clear();
      }
    }
    
    /*
     *  Create arcs
     */
    for (int i=0; i < size; i++) {
      SimulatedNest nest = (SimulatedNest) nests.get(i);
      addConnections(nest, i);
    }
  }

  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods from Peernet
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the nodes composing the simulated p2p network
   */
  public List getNests()
  {
    return nests;
  }

  /**
   *  Returns the initial size of the simulated p2p network (number of nests)
   *
   *  @return size of the p2p network.
   */
  public int size()
  {
    return size;
  }

  /**
   *  Returns the initial degree of the simulated p2p network (number of nests)
   *
   *  @return degree of the p2p network.
   */
  public int degree()
  {
    return degree;
  }

  
  /*
   *  FIXME: How the other nodes discover about
   *  the new hosts? The problem is that ants generated
   *  from this node should be able to advertise the
   *  existance of a copy there. For example, Insant
   *  are necessary. Or, Gnutant should try to search
   *  also randomly (not following the best value).
   */
  /** 
   *  Executes a round in each of the nodes constituting
   *  the peernet.
   */
  public void executeRound()
  {

    /* If the leaveProb is greater than 0, remove nests */
    if (leaveProb > 0.0)
      removeNests();
    
    /* If the joinProb is greater than 0, add new nests */
    if (joinProb > 0.0)
      addNests();

    /* Each nest executes the code of an ant in its queue */
    for (int j=0; j < nests.size(); j++) {
      SimulatedNest nest = (SimulatedNest) nests.get(j);
      nest.getManager().execute();
    }

  }

  /**
   *  Remove nests from the simulated peer network, depending
   *  on the leave probability.
   */
  private void removeNests()
  {
    /* Removes crashed/leaving members */
    int toRemove = (int) (nests.size() * leaveProb);

    //FIXME; This will remove a constant number of nests for each invocation.
    // We should use binomal distribution to compute how many to remove instead
    // of the approach used below and also used in the addNests() method.

    for (int index = 0; toRemove > 0; toRemove--) {
      index = random.nextInt(nests.size());
      SimulatedNest nest = (SimulatedNest) nests.get(index);
      nest.leave();
      /*
       * If there are no new nests to be added, substitute with the
       * last one.
       */
      nests.set(index, nests.get(nests.size()-1));
      nests.remove(nests.size()-1);
      size = nests.size();
      Log.println("Removed nest: "+ nest.getNestId() + ";\t  new size: " +size);
    }

    /*
    for (int i=0; i < nests.size(); i++) {
    float rnd = random.nextFloat();
    if (rnd < leaveProb) {
    SimulatedNest nest = (SimulatedNest) nests.get(i);
    nest.leave();
    nests.set(i, nests.get(nests.size()-1));
    nests.remove(nests.size()-1);
    size = nests.size();
    Log.println("Removed nest: "+ nest.getNestId() + ";\t  new size: " +size);
    }
    }
     */
  }


  /**
   *  Add new nests to the simulated peer network, 
   *  depending on the join probability. 
   */
  private void addNests()
  {
    /* Add new members */
    for (int i=0; i < nests.size(); i++) {
      float rnd = random.nextFloat();
      if (rnd < joinProb) {
        /* Add a new nest to the end of the list */
        SimulatedNest nest = (SimulatedNest) factory.get(SimulatedNest.NAME);
        nests.add(nest);
        size = nests.size();
        Log.println("Added   nest: "+ nest.getNestId() + ";\t  new size: " +size);
        addConnections(nest, size);
      }
    }
  }
    
  /**
   *  Add connections to the existing network.
   */
  private void addConnections(SimulatedNest nest, int index)
  {
    /* Add connections to nests */
    for (int i=0; i < degree; i++) {
      int rndIndex = index;
      while (rndIndex == index)  
        rndIndex = random.nextInt(size);
      SimulatedNest rndNest = (SimulatedNest) nests.get(rndIndex);
      rndNest.addClientNeighbour(nest.getNestId());
      nest.addClientNeighbour(rndNest.getNestId());
      //Log.println("NEIGHBOUR: " + nest.getId() + " " + rndNest.getId());
    } 
  }
  
} // END PeernetImpl

