/*
 * Copyright (c) 2001 The Anthill Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

package antsim.impl;

import java.util.*;

import anthill.factory.Factory;
import anthill.util.Clearable;
import antsim.ResourceSet;

/**
 *  <CODE>RandomResourceSet</CODE> is a simple implementation of <CODE>ResourceSet</CODE>
 *  which generates <CODE>RandomResource</CODE>s. Each random resource has a generic
 *  name, which is the string representing a random long value. Requests are obtained
 *  by selecting randomly one of the resources stored in this <CODE>ResourceSet</CODE>
 *  and by returning the string containing the resource name.
 *  
 *  @author Alberto Montresor
 *  @version $Revision: 1.3 $
 */
public class RandomResourceSet
implements ResourceSet, Clearable
{
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Static section
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Random generator */
  private static Random random = new Random();

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameter identifiers
  ////////////////////////////////////////////////////////////////////////////////////////////
  
  /**
   *  String name of the parameter used to determine the number of
   *  of resources contained in this resource set.
   */
  public static final String PAR_SIZE = "Size";

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Parameters
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Number of resources contained in this resource set */
  private int size;
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Fields
  ////////////////////////////////////////////////////////////////////////////////////////////

  /** Factory object used to read configuration parameters and classes */
  private Factory factory;
  
  /** Resources contained in this resource set */
  private ArrayList resources;

  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Constructor and clear() method
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Initializes a new random resource set. Configuration parameters are
   *  read through the 
   *  
   *  @param factory the factory to be used to read configuration arguments
   *    and to create other configurable objects
   *  @param name the name that is associated to this instance in the
   *    configuration file
   */
  public RandomResourceSet(Factory factory, String name)
  {
    this.factory = factory;
    clear();
  }
    
  /**
   *  Restores the object to its initial state.
   */
  public void clear()
  {
    size = factory.intValue(NAME + "." +PAR_SIZE);
    if (resources == null)
      resources = new ArrayList(size);
    
    while (size > resources.size()) {
      // Missing resources; we insert new resources
      resources.add(new RandomResource());
    }
      
    while (size < resources.size()) {
      // Too much resources; we remove them
      resources.remove(0);
    }
  }
  
  
  ////////////////////////////////////////////////////////////////////////////////////////////
  // Methods
  ////////////////////////////////////////////////////////////////////////////////////////////

  /**
   *  Return the number of resources generated and stored in
   *  this <CODE>ResourceSet</CODE>.
   */
  public int size()
  {
    return size;
  }

  /**
   *  Return an array containing the resources generated and 
   *  stored in this <CODE>ResourceSet</CODE>.
   */
  public List getResources()
  {
    return resources;
  }
  
} // END RandomResourceSet

