/******************************************************
 * File: Dowloader.java
 * created 06-Mar-02 12:12:32 PM by root
 */


package jxtaimpl;

import java.io.IOException;
import java.net.*;

import jxtaimpl.storage.*;

import net.jxta.document.AdvertisementFactory;
import net.jxta.endpoint.Message;
import net.jxta.id.IDFactory;
import net.jxta.pipe.*;
import net.jxta.protocol.PipeAdvertisement;

import anthill.storage.CapacityException;

/**
*  This class is in charge of downloading resources found by ants. This is accomplished
*  getting connected to the <code>DownloadService</code> deamon whose pipe advertisement
*  can be created starting from the <code>pipeXXX</code> properties found in the <code>
*  AnthillResource</code> instance 
*  @author Francesco Russo (frusso@cs.unibo.it)
*/
public class Downloader extends Thread
{
	/**
	*  This is this class' constructor. It needs an <code>AnthillResource</code> describing
	*  the resource to download, a reference to the <code>AnthillGroup</code> and to the local
	*  <code>CmsStorage</code> where the resource will be saved
	*
	* @param res The <code>AnthillResource</code> describing the resource to download
	* @param grp A reference to the <code>AnthillGroup</code> instance
	* @param storage A reference to the <code>jxtaimpl.storage.CmsStorage</code> instance
	*				 were the downloaded resource should be placed
	*/
	public Downloader(AnthillResource res, AnthillGroup grp, CmsStorage storage)
	{
		try
		{
			resource = res;
			anthillGroup = grp;
			this.storage = storage;
			pipeService = anthillGroup.getNetPipeService();
		
			outPipeAdv = (PipeAdvertisement)AdvertisementFactory.newAdvertisement( PipeAdvertisement.getAdvertisementType() );
	 		outPipeAdv.setName( resource.getPipeName() );
			URL url = resource.getPipeId();
			outPipeAdv.setPipeID( (PipeID)IDFactory.fromURL(url) );
			outPipeAdv.setType( resource.getPipeType() );
			
			inPipeAdv = (PipeAdvertisement)AdvertisementFactory.newAdvertisement( PipeAdvertisement.getAdvertisementType() );
	 		inPipeAdv.setName( "tempPipe" );
			inPipeAdv.setPipeID( IDFactory.newPipeID(anthillGroup.getPeerGroupID()) );
			inPipeAdv.setType( PipeService.UnicastType );
		}
		catch(MalformedURLException mue)
		{
			mue.printStackTrace();
		}
		catch(UnknownServiceException use)
		{
			use.printStackTrace();
		}
	}
	
	/**
	*  This method implements the Downloader's logic
	*/
	public void run()
	{                	
		try
		{
			outPipe = pipeService.createOutputPipe( outPipeAdv, WAITTIME );
			inPipe = pipeService.createInputPipe( inPipeAdv );
			
			Message outMsg = pipeService.createMessage();
			
			String pipeIdURL = inPipeAdv.getPipeID().getURL().toString();
			outMsg.setString( PIPEID, pipeIdURL );
			outMsg.setString( PIPENAME, inPipeAdv.getName() );
			outMsg.setString( PIPETYPE, inPipeAdv.getType() );
			outMsg.setString( RESOURCEID, resource.getKey().toString() );
			outMsg.setString( SPECIES, resource.getSpecies() );
			
			outPipe.send( outMsg );
			
			Message inMsg = inPipe.waitForMessage();
			
			byte[] resContent = inMsg.getBytes( RESOURCE );
			try
			{
				storage.addResource( null, resource, resContent );
			}
			catch(CapacityException ce)
			{
				ce.printStackTrace();
			}
		}
		catch(IOException ioe)
		{
			ioe.printStackTrace();
		}
		catch(InterruptedException ie)
		{
			ie.printStackTrace();
		}
	}
	
	private AnthillGroup anthillGroup;
	private CmsStorage storage;
	private AnthillResource resource;
	private PipeService pipeService;
	private PipeAdvertisement outPipeAdv;
	private PipeAdvertisement inPipeAdv;
	private OutputPipe outPipe;
	private InputPipe inPipe;
	
	private static final long WAITTIME = 10000;
	
	//* tags for incoming requests for downloading resources
	private static final String PIPEID = "PIPEID";
	private static final String PIPENAME = "PIPENAME";
	private static final String PIPETYPE = "PIPETYPE";
	private static final String RESOURCEID = "RESOURCEID";
	private static final String SPECIES = "SPECIES";
	
	//* tag for outgoing messages for uploading resources
	private static final String RESOURCE = "RESOURCE";
}
