/******************************************************
 * File: AnthillResource.java
 * created 13-Jan-02 4:20:53 PM by frusso@cs.unibo.it
 */

package jxtaimpl.storage;

import java.io.*;
import java.lang.reflect.*;
import java.net.URL;

import net.jxta.share.*;

import anthill.Resource;

/**
 * This class represents a resource in the Anthill network
 * @author Francesco Russo (frusso@cs.unibo.it) 
 */
public class AnthillResource implements Resource, Serializable
{
  /**
   * This is the default constructor.
   * It needs a Content instance, and the set of information rapresenting the input pipe
   * from which a resource can be downloaded
   *
   * @param cont The <code>Content</code> instance associate with the resource
   * @param species The ant species of the ant requiring this resource
   * @param pipeId This is the JXTA pipeID of the local <code>jxtaimpl.DownloadService</code>
   *               input-pipe waiting for download requests to come
   * @param pipeName The pipe's logical name
   * @param pipeType The pipe's type (point-to-point, propagate, secure)
   */
  public AnthillResource( Content cont, String species, URL pipeId, String pipeName, String pipeType )
  {
	super();
	this.setContent( cont );
	this.setSpecies( species );
	this.pipeId = pipeId;
	this.pipeName = pipeName;
	this.pipeType = pipeType;
	
	//System.out.println("--- AnthillResource: "+pipeId.toString()+" "+pipeName+" "+pipeType);
  }
  

  /**
  * Returns the key identifier of the object.
  * Actually this key idendifier is a <code>net.jxta.share.ContentId</code> instance.
  */
  public Object getKey()
  {
	ContentAdvertisement adv = content.getContentAdvertisement();
	return adv.getContentId();
  }
  
  /**
  * Returns the string identifier of the object.
  * Actually this string idendifier is a <code>net.jxta.share.ContentId</code> instance
  * converted to String.
  */
  public String getName()
  {
	ContentAdvertisement adv = content.getContentAdvertisement();
	return adv.getContentId().toString();
  }
  
  /**
  * This method returns the logical name of the resource.
  * In order to obtain its ID, please refer to the <CODE>getKey</CODE>
  * and <CODE>getName</CODE> methods.
  */
  public String getResourceName()
  {
	ContentAdvertisement adv = content.getContentAdvertisement();
	return adv.getName();
  }
	
  /**
  *  Returns the value associated to the specified attribute
  *  name, or null if there is no attribute with that name.
  *  The attributes this implementation refers to are the elements
  *  found in a generic ContentAdvertisement, since an <code>AnthillResource</code>
  *  is based on the <code>net.jxta.share.Content</code> class. Reflection API is used
  *  for discovering getter methods in the ContentAdvertisement class matching
  *  with the "name" input parameter.
  *
  *  @param name the name of the attribute (ContentAdvertisement xml-element)
  */
  public String getAttribute(String name)
  {
	try
	{
	  ContentAdvertisement adv = content.getContentAdvertisement();
	  Class contAdvClass = Class.forName("net.jxta.share.ContentAdvertisement");
	  Method[] methods = contAdvClass.getMethods();
	  int i = 0;
	
	  for (i = 0; i < methods.length; i++)
	  {
		if ( methods[i].getName().startsWith("get") )
		{
		  if ( name.toLowerCase().equals( methods[i].getName().substring(3).toLowerCase() ) )
		  {
			Object obj = methods[i].invoke( adv, null );
			return obj.toString();
		  }
		}
	  }
	}
	catch(ClassNotFoundException cnfe)
	{
	  cnfe.printStackTrace();
	}
	catch(IllegalAccessException iae)
	{
	  iae.printStackTrace();
	}
	catch(InvocationTargetException ite)
	{
	  ite.printStackTrace();
	}
	return null;
  }
  
	
  /**
  * This method returns an InputStream created starting from the
  * <code>Content</code> this resource refers to. This method has to
  * be invoke only locally.
  *
  *  @return an input stream to read the contents of a resource;
  */
  public InputStream getInputStream() throws IOException
  {
	//ByteArrayInputStream bais = new ByteArrayInputStream( resource );
	//return bais;
	return content.getInputStream();
  }
	
  /**
  *  Return the lenght of the content of the resource, if
  *  available; otherwise, it returns -1.
  */
  public long getContentLength()
  {
	//return resource.length;
	ContentAdvertisement adv = content.getContentAdvertisement();
	return adv.getLength();
  }
  
  
  /**
  *  Return the space required to store the resource object.
  */
  public long size()
  {
	//return resource.length;
	return this.getContentLength();
  }

  /**
   * Method invoked by the constructor in order to properly
   * set the Content instance this object has to wrap.
   *
   * @param cont The <code>Content</code> instance
   */
  private void setContent(Content cont)
  {
	//int BUFFER_SIZE = 0;
	
	//try
	//{
	  content = cont;
	  //InputStream is = content.getInputStream();
	  //BUFFER_SIZE = is.available();
	  //byte[] tempBuffer = new byte[ BUFFER_SIZE ];
	  //ByteArrayOutputStream baos = new ByteArrayOutputStream();
	
	  //while( is.read( tempBuffer ) != -1  )
	  //{
		//baos.write( tempBuffer );
	  //}
	  
	  //is.close();
	  //resource = baos.toByteArray();
	  //baos.close();
	//}
	//catch(IOException ioe)
	//{
	  //ioe.printStackTrace();
	//}
  }
  
  
  /**
  * This protected method can be invoked in order to obtain the Content
  * instance associated with an AnthillResource.
  */
  protected Content getContent()
  {
	return content;
  }
  
  /**
  *  This method sets the ant species of the ant that obtained this resource
  */
  private void setSpecies(String species)
  {
	this.species = species;
  }

  /**
  *  This method gets the ant species of the ant that obtained this resource
  */
  public String getSpecies()
  {
	return this.species;
  }
  
  /**
  *  This method returns the ID of the pipe for downloading the resource
  */
  public URL getPipeId()
  {
  	return pipeId;
  }
  
  /**
  *  This method returns the name of the pipe for downloading the resource
  */
  public String getPipeName()
  {
  	return pipeName;
  }
 
  /**
  *  This method returns the type of the pipe for downloading the resource
  */
  public String getPipeType()
  {
  	return pipeType;
  }
  
  
  //* fields specifieing the availabel input pipe for downloading the resource
  private URL pipeId;
  private String pipeName;
  private String pipeType;
  
  private String species;

  /**
   * This is a descriptor for the resource. 
   */
  private Content content;

  /**
   * This is the real resource. This byte array is filled in invoking the
   * "setContent(Content cont)" method.
   */
  //private byte[] resource;
}
